from collections import defaultdict
from functools import reduce

from titus_isolate.model.processor import utils


class Cpu:
    def __init__(self, packages):
        if len(packages) < 1:
            raise ValueError("A CPU must contain at least 1 package.")

        self.__packages = packages

    def get_packages(self):
        return self.__packages

    def get_emptiest_package(self):
        emptiest_package = self.__packages[0]
        curr_empty_thread_count = len(emptiest_package.get_empty_threads())

        for package in self.get_packages()[1:]:
            new_empty_thread_count = len(package.get_empty_threads())

            if new_empty_thread_count > curr_empty_thread_count:
                emptiest_package = package
                curr_empty_thread_count = new_empty_thread_count

        return emptiest_package

    def get_cores(self):
        return reduce(list.__add__, [package.get_cores() for package in self.get_packages()])

    def get_threads(self):
        return reduce(list.__add__, [package.get_threads() for package in self.get_packages()])

    def get_empty_threads(self):
        return utils.get_empty_threads(self.get_threads())

    def get_claimed_threads(self):
        return utils.get_claimed_threads(self.get_threads())

    def clear(self):
        for t in self.get_threads():
            t.clear()

    def get_workload_ids_to_thread_ids(self):
        res = defaultdict(list)
        for t in self.get_threads():
            if t.is_claimed():
                for w_id in t.get_workload_ids():
                    res[w_id] += [t.get_id()]
        return res

    def get_natural_indexing_2_original_indexing(self):
        return {i: t.get_id() for i, t in enumerate(self.get_threads())}

    def to_dict(self):
        packages = []
        for p in self.get_packages():

            cores = []
            for c in p.get_cores():

                threads = []
                for t in c.get_threads():
                    threads.append({
                        "id": t.get_id(),
                        "workload_id": t.get_workload_ids()
                    })
                cores.append({
                    "id": c.get_id(),
                    "threads": threads
                })

            packages.append({
                "id": p.get_id(),
                "cores": cores
            })

        return {
            "packages": packages
        }

    def __str__(self):
        n_packages = len(self.get_packages())
        n_cores = n_packages * len(self.get_packages()[0].get_cores())
        return ('%i packages, %i cores per package, %i threads per core\n'
            '%i threads claimed\n'
            '%s') % (
                n_packages,
                n_cores,
                int(len(self.get_threads()) / n_cores),
                len(self.get_claimed_threads()),
                utils.visualize(self)
            )



import logging
import os

log = logging.getLogger()
log.setLevel(logging.INFO)

LOG_FMT_STRING = '%(asctime)s,%(msecs)d %(levelname)s [%(filename)s:%(lineno)d] %(message)s'


if "DISTRIB_ID" in os.environ:
    from systemd.journal import JournaldLogHandler
    journald_handler = JournaldLogHandler()
    journald_handler.setFormatter(logging.Formatter(LOG_FMT_STRING))
    log.addHandler(journald_handler)



import platform
import subprocess

from titus_isolate.model.processor.core import Core
from titus_isolate.model.processor.cpu import Cpu
from titus_isolate.model.processor.package import Package
from titus_isolate.model.processor.thread import Thread
from titus_isolate.model.processor.utils import DEFAULT_PACKAGE_COUNT, DEFAULT_CORE_COUNT, DEFAULT_THREAD_COUNT


def get_cpu_from_env():
    system = platform.system()
    processor = None

    if system == 'Darwin':
        processor = MacProcessor()
    elif system == 'Linux':
        processor = LinuxProcessor()

    if processor is None:
        raise EnvironmentError("Unexpected system type: '{}'".format(system))

    return get_cpu(processor.get_package_count(), processor.get_cores_per_package(), processor.get_threads_per_core())


def get_cpu(
        package_count=DEFAULT_PACKAGE_COUNT,
        cores_per_package=DEFAULT_CORE_COUNT,
        threads_per_core=DEFAULT_THREAD_COUNT):
    packages = []
    for p_i in range(package_count):

        cores = []
        for c_i in range(cores_per_package):
            cores.append(
                Core(c_i, __get_threads(p_i, c_i, package_count, cores_per_package, threads_per_core)))

        packages.append(Package(p_i, cores))

    return Cpu(packages)


def __get_threads(package_index, core_index, package_count, core_count, thread_count):
    threads = []
    for row_index in range(thread_count):
        offset = row_index * package_count * core_count
        index = offset + package_index * core_count + core_index
        threads.append(Thread(index))

    return threads


class Processor:
    def __init__(self, package_count, cores_per_package, threads_per_core):
        self.__package_count = package_count
        self.__cores_per_package = cores_per_package
        self.__threads_per_core = threads_per_core

    def get_package_count(self):
        return self.__package_count

    def get_cores_per_package(self):
        return self.__cores_per_package

    def get_threads_per_core(self):
        return self.__threads_per_core


class LinuxProcessor(Processor):
    def __init__(self):
        super(LinuxProcessor, self).__init__(
            self.__get_package_count(),
            self.__get_cores_per_package(),
            self.__get_threads_per_core())

    def __get_package_count(self):
        return self.__get_lscpu_value('^Socket')

    def __get_cores_per_package(self):
        return self.__get_lscpu_value('^Core')

    def __get_threads_per_core(self):
        return self.__get_lscpu_value('^Thread')

    @staticmethod
    def __get_lscpu_value(grep_filter):
        output = subprocess.check_output("lscpu | grep -E {}".format(grep_filter), shell=True).decode('utf-8')
        return int(output.split(':')[1].strip())


class MacProcessor(Processor):
    def __init__(self):
        super(MacProcessor, self).__init__(
            self.__get_package_count(),
            self.__get_cores_per_package(),
            self.__get_threads_per_core())

    def __get_package_count(self):
        return self.__get_sysctl_value('hw.packages')

    def __get_cores_per_package(self):
        return int(self.__get_physical_cores() / self.__get_package_count())

    def __get_threads_per_core(self):
        logical_cpus = self.__get_sysctl_value('hw.logicalcpu')
        return int(logical_cpus / self.__get_physical_cores())

    def __get_physical_cores(self):
        return self.__get_sysctl_value('hw.physicalcpu')

    @staticmethod
    def __get_sysctl_value(key):
        return int(subprocess.check_output(['sysctl', '-n', key]).decode('utf-8').strip())



from titus_isolate.config.agent_property_provider import AgentPropertyProvider


class ConfigManager:

    def __init__(self, property_provider=AgentPropertyProvider()):
        self.__property_provider = property_provider

    def get(self, key, default=None):
        value = self.__property_provider.get(key)

        if value is None:
            return default
        else:
            return value

    def get_region(self):
        return self.__property_provider.get('EC2_REGION')

    def get_environment(self):
        return self.__property_provider.get('NETFLIX_ENVIRONMENT')



from abc import abstractmethod

from titus_isolate.metrics.metrics_reporter import MetricsReporter


class CgroupManager(MetricsReporter):

    @abstractmethod
    def set_cpuset(self, container_name, thread_ids):
        pass

    @abstractmethod
    def get_cpuset(self, container_name):
        pass

    @abstractmethod
    def release_cpuset(self, container_name):
        pass

    @abstractmethod
    def get_isolated_workload_ids(self):
        pass



import os
import time

from titus_isolate.exit_handler import ExitHandler


class RealExitHandler(ExitHandler):

    def exit(self, code):
        # Sleep for 1 second so log messages get flushed.
        time.sleep(1)
        os._exit(code)



import requests

from titus_isolate.allocate.cpu_allocate_exception import CpuAllocationException
from titus_isolate.allocate.cpu_allocator import CpuAllocator
from titus_isolate.allocate.utils import get_threads_body, parse_cpu, get_rebalance_body
from titus_isolate.config.constants import REMOTE_ALLOCATOR_URL
from titus_isolate.model.processor.cpu import Cpu
from titus_isolate.utils import get_config_manager


class RemoteCpuAllocator(CpuAllocator):
    def __init__(self, free_thread_provider):
        self.__url = get_config_manager().get(REMOTE_ALLOCATOR_URL, "http://localhost:7501")
        self.__headers = {'Content-Type': "application/json"}
        self.__reg = None

    def assign_threads(self, cpu: Cpu, workload_id: str, workloads: dict) -> Cpu:
        url = "{}/assign_threads".format(self.__url)
        body = get_threads_body(cpu, workload_id, workloads)
        response = requests.put(url, json=body, headers=self.__headers)

        if response.status_code == 200:
            return parse_cpu(response.json())

        raise CpuAllocationException("Failed to assign threads: {}".format(response.text))

    def free_threads(self, cpu: Cpu, workload_id: str, workloads: dict) -> Cpu:
        url = "{}/free_threads".format(self.__url)
        body = get_threads_body(cpu, workload_id, workloads)
        response = requests.put(url, json=body, headers=self.__headers)

        if response.status_code == 200:
            return parse_cpu(response.json())

        raise CpuAllocationException("Failed to free threads: {}".format(response.text))

    def rebalance(self, cpu: Cpu, workloads: dict) -> Cpu:
        url = "{}/rebalance".format(self.__url)
        body = get_rebalance_body(cpu, workloads)
        response = requests.put(url, json=body, headers=self.__headers)

        if response.status_code == 200:
            return parse_cpu(response.json())

        raise CpuAllocationException("Failed to rebalance threads: {}".format(response.text))

    def set_registry(self, registry):
        pass

    def report_metrics(self, tags):
        pass



from titus_isolate import log
from titus_isolate.model.processor.utils import get_workload_ids


def get_updates(cur_cpu, new_cpu):
    updates = {}
    for workload_id in get_workload_ids(new_cpu):
        new_thread_ids = __get_threads(new_cpu, workload_id)
        cur_thread_ids = __get_threads(cur_cpu, workload_id)
        if set(new_thread_ids) != set(cur_thread_ids):
            log.info("workload: '{}' updated threads from: '{}' to: '{}'".format(workload_id, cur_thread_ids, new_thread_ids))
            updates[workload_id] = new_thread_ids

    return updates


def __get_threads(cpu, workload_id):
    return [t.get_id() for t in cpu.get_threads() if workload_id in t.get_workload_ids()]



from titus_isolate.event.constants import ACTION, ACTOR, ATTRIBUTES, CREATE, REQUIRED_LABELS
from titus_isolate.event.event_handler import EventHandler
from titus_isolate.model.utils import get_workload_from_event


class CreateEventHandler(EventHandler):
    def __init__(self, workload_manager):
        super().__init__(workload_manager)

    def handle(self, event):
        if not self.__relevant(event):
            return

        workload = get_workload_from_event(event)

        self.handling_event(event, "adding workload: '{}'".format(workload.get_id()))
        self.workload_manager.add_workload(workload)
        self.handled_event(event, "added workload: '{}'".format(workload.get_id()))

    def __relevant(self, event):
        if not event[ACTION] == CREATE:
            self.ignored_event(event, "not a CREATE event")
            return False

        for expected_label in REQUIRED_LABELS:
            if expected_label not in event[ACTOR][ATTRIBUTES]:
                self.ignored_event(event, "container created without expected label: '{}'".format(expected_label))
                return False

        return True



import schedule

from titus_isolate import log
from titus_isolate.constants import GENERIC_PROPERTY_CHANGE_EXIT, ALLOCATOR_CONFIG_CHANGE_EXIT
from titus_isolate.isolate.utils import get_allocator

PROPERTY_CHANGE_DETECTION_INTERVAL_SEC = 10


class RestartPropertyWatcher:

    def __init__(
            self,
            config_manager,
            exit_handler,
            properties,
            detection_interval=PROPERTY_CHANGE_DETECTION_INTERVAL_SEC):

        self.__config_manager = config_manager
        self.__exit_handler = exit_handler
        self.__properties = properties

        self.__original_properties = {}
        for p in properties:
            self.__original_properties[p] = config_manager.get(p)

        self.__original_primary_allocator_name =\
            get_allocator(config_manager).get_primary_allocator().__class__.__name__

        log.info("Starting watching for changes to properties: {}".format(properties))
        for k, v in self.__original_properties.items():
            log.info("{}: {}".format(k, v))

        schedule.every(detection_interval).seconds.do(self.detect_changes)

    def detect_changes(self):
        self.__detect_property_changes()
        self.__detect_ab_changes()

    def __detect_property_changes(self):
        for p in self.__properties:
            original_value = self.__original_properties[p]
            curr_value = self.__config_manager.get(p)
            log.debug("property: '{}' original: '{}' current: '{}'".format(p, original_value, curr_value))

            if original_value != curr_value:
                log.info("Restarting because property: '{}' changed from: '{}' to: '{}'".format(
                    p, original_value, curr_value))

                self.__exit_handler.exit(GENERIC_PROPERTY_CHANGE_EXIT)

    def __detect_ab_changes(self):
        curr_primary_allocator_name = get_allocator(self.__config_manager).get_primary_allocator().__class__.__name__
        if self.__original_primary_allocator_name != curr_primary_allocator_name:
            log.info("Restarting because primary CPU allocator changed from: '{}' to: '{}'".format(
                self.__original_primary_allocator_name, curr_primary_allocator_name))
            self.__exit_handler.exit(ALLOCATOR_CONFIG_CHANGE_EXIT)



from titus_isolate import log
from titus_isolate.cgroup.cgroup_manager import CgroupManager
from titus_isolate.constants import RECONCILIATION_FAILURE_EXIT
from titus_isolate.exit_handler import ExitHandler
from titus_isolate.model.processor.cpu import Cpu


class Reconciler:

    def __init__(self, cgroup_manager: CgroupManager, exit_handler: ExitHandler):
        self.__cgroup_manager = cgroup_manager
        self.__exit_handler = exit_handler
        self.__unisolated_workload_count = 0
        self.__missing_cpuset_count = 0

    def reconcile(self, cpu: Cpu):
        workloads = self.get_workloads(cpu)
        for w_id, t_ids in workloads.items():
            isolated_workload_ids = self.__cgroup_manager.get_isolated_workload_ids()
            if w_id not in isolated_workload_ids:
                log.warn("Workload: '{}' is not in isolated set: '{}' so unable to retrieve cpuset during reconciliation.".format(w_id, isolated_workload_ids))
                self.__unisolated_workload_count += 1
                continue

            cpuset = self.__cgroup_manager.get_cpuset(w_id)
            if cpuset is None:
                log.warn("Workload likely exited during reconciliation: '{}'".format(w_id))
                self.__missing_cpuset_count += 1
                continue

            cpuset = sorted(cpuset)
            t_ids = sorted(t_ids)

            if cpuset != t_ids:
                log.error("Reconciliation has failed for workload: '{}', cpuset: {} != t_ids: {}".format(
                    w_id, cpuset, t_ids))
                self.__exit_handler.exit(RECONCILIATION_FAILURE_EXIT)
            else:
                log.info("Reconciliation has succeeded for workload: '{}', cpuset: {} == t_ids: {}".format(
                    w_id, cpuset, t_ids))

    def get_unisolated_workload_count(self):
        return self.__unisolated_workload_count

    def get_missing_cpuset_count(self):
        return self.__missing_cpuset_count

    def get_total_warning_count(self):
        return self.get_unisolated_workload_count() + self.get_missing_cpuset_count()

    @staticmethod
    def get_workloads(cpu: Cpu):
        workloads = {}
        for t in cpu.get_threads():
            for w_id in t.get_workload_ids():
                if w_id in workloads:
                    workloads[w_id].append(t.get_id())
                else:
                    workloads[w_id] = [t.get_id()]

        return workloads



from titus_isolate import log
from titus_isolate.allocate.cpu_allocator import CpuAllocator
from titus_isolate.cgroup.file_cgroup_manager import FileCgroupManager
from titus_isolate.model.processor.cpu import Cpu


class NoopResetCpuAllocator(CpuAllocator):

    def __init__(self, free_thread_provider="", cgroup_manager=FileCgroupManager()):
        self.__cgroup_manager = cgroup_manager

    def get_cgroup_manager(self):
        return self.__cgroup_manager

    def assign_threads(self, cpu, workload):
        thread_count = len(cpu.get_threads())
        thread_ids = list(range(thread_count))

        log.info("Setting cpuset.cpus to ALL cpus: '{}' for workload: '{}'".format(thread_ids, workload.get_id()))
        self.__cgroup_manager.set_cpuset(workload.get_id(), thread_ids)

        return cpu

    def free_threads(self, workload_id):
        log.info("Ignoring attempt to free threads for workload: '{}'".format(workload_id))

    def rebalance(self, cpu: Cpu, workloads: dict) -> Cpu:
        return cpu

    def set_registry(self, registry):
        pass

    def report_metrics(self, tags):
        pass



from titus_isolate import log
from titus_isolate.config.constants import TITUS_ISOLATE_DYNAMIC_NUMA_BALANCING, \
    DEFAULT_TITUS_ISOLATE_DYNAMIC_NUMA_BALANCING
from titus_isolate.model.processor.cpu import Cpu
from titus_isolate.model.workload import Workload
from titus_isolate.utils import get_config_manager

NUMA_BALANCING_PATH = '/proc/sys/kernel/numa_balancing'


def update_numa_balancing(workload: Workload, cpu: Cpu):
    try:
        config_manager = get_config_manager()
        dynamic_numa_balancing_enabled = bool(config_manager.get(
            TITUS_ISOLATE_DYNAMIC_NUMA_BALANCING,
            DEFAULT_TITUS_ISOLATE_DYNAMIC_NUMA_BALANCING))

        if not dynamic_numa_balancing_enabled:
            enable_numa_balancing()
            return

        if _occupies_entire_cpu(workload, cpu):
            disable_numa_balancing()
        else:
            enable_numa_balancing()
    except:
        log.exception("Failed to update NUMA balancing.")


def enable_numa_balancing():
    if _get_numa_balancing() == 1:
        log.info("NUMA balancing is enabled.")
        return

    log.info("Enabling NUMA balancing.")
    _set_numa_balancing(1)


def disable_numa_balancing():
    if _get_numa_balancing() == 0:
        log.info("NUMA balancing is disabled.")
        return

    log.info("Disabling NUMA balancing")
    _set_numa_balancing(0)


def _get_numa_balancing():
    with open(NUMA_BALANCING_PATH, 'r') as f:
        return int(f.read())


def _set_numa_balancing(state: int):
    if state != 0 and state != 1:
        raise ValueError("numa_balancing may be set to either 0 or 1, received: '{}'".format(state))

    with open(NUMA_BALANCING_PATH, 'w') as f:
        f.write(str(state))


def _occupies_entire_cpu(workload: Workload, cpu: Cpu):
    return len(cpu.get_threads()) == workload.get_thread_count()



from datetime import datetime as dt
from collections import defaultdict
import copy
from math import ceil, floor

from titus_optimize.compute_v2 import IP_SOLUTION_TIME_BOUND, optimize_ip, IPSolverParameters

from titus_isolate import log
from titus_isolate.allocate.cpu_allocator import CpuAllocator
from titus_isolate.config.config_manager import ConfigManager
from titus_isolate.config.constants import ALPHA_NU, DEFAULT_ALPHA_NU, ALPHA_LLC, DEFAULT_ALPHA_LLC, ALPHA_L12, \
    DEFAULT_ALPHA_L12, ALPHA_ORDER, DEFAULT_ALPHA_ORDER, ALPHA_PREV, DEFAULT_ALPHA_PREV, BURST_MULTIPLIER, \
    DEFAULT_BURST_MULTIPLIER, MAX_BURST_POOL_INCREASE_RATIO, DEFAULT_MAX_BURST_POOL_INCREASE_RATIO, \
    BURST_CORE_COLLOC_USAGE_THRESH, DEFAULT_BURST_CORE_COLLOC_USAGE_THRESH, WEIGHT_CPU_USE_BURST, \
    DEFAULT_WEIGHT_CPU_USE_BURST
from titus_isolate.event.constants import BURST, STATIC
from titus_isolate.metrics.constants import IP_ALLOCATOR_TIMEBOUND_COUNT, FORECAST_REBALANCE_FAILURE_COUNT
from titus_isolate.model.processor.cpu import Cpu
from titus_isolate.model.utils import get_burst_workloads, release_all_threads
from titus_isolate.model.utils import get_sorted_workloads
from titus_isolate.monitor.workload_monitor_manager import WorkloadMonitorManager
from titus_isolate.predict.cpu_usage_predictor import PredEnvironment
from titus_isolate.predict.cpu_usage_predictor_manager import CpuUsagePredictorManager


class CUVector:
    def __init__(
            self,
            requested_cus,
            curr_placement_vectors_static,
            ordered_workload_ids_static,
            ordered_workload_ids_burst,
            burst_pool_size_req):
        self.requested_cus = requested_cus
        self.curr_placement_vectors_static = curr_placement_vectors_static
        self.ordered_workload_ids_static = ordered_workload_ids_static
        self.ordered_workload_ids_burst = ordered_workload_ids_burst
        self.burst_pool_size_req = burst_pool_size_req

    def __str__(self):
        return str(vars(self))


class ForecastIPCpuAllocator(CpuAllocator):

    def __init__(self,
                 cpu_usage_predictor_manager: CpuUsagePredictorManager,
                 config_manager: ConfigManager,
                 workload_monitor_manager: WorkloadMonitorManager,
                 solver_max_runtime_secs: int = 5):
        self.__reg = None
        self.__time_bound_call_count = 0
        self.__rebalance_failure_count = 0
        self.__ip_solver_params = IPSolverParameters(
            alpha_nu=config_manager.get(ALPHA_NU, DEFAULT_ALPHA_NU),
            alpha_llc=config_manager.get(ALPHA_LLC, DEFAULT_ALPHA_LLC),
            alpha_l12=config_manager.get(ALPHA_L12, DEFAULT_ALPHA_L12),
            alpha_order=config_manager.get(ALPHA_ORDER, DEFAULT_ALPHA_ORDER),
            alpha_prev=config_manager.get(ALPHA_PREV, DEFAULT_ALPHA_PREV),
            burst_multiplier=config_manager.get(BURST_MULTIPLIER, DEFAULT_BURST_MULTIPLIER),
            max_burst_pool_increase_ratio=config_manager.get(MAX_BURST_POOL_INCREASE_RATIO, DEFAULT_MAX_BURST_POOL_INCREASE_RATIO),
            burst_core_colloc_usage_thresh=config_manager.get(BURST_CORE_COLLOC_USAGE_THRESH, DEFAULT_BURST_CORE_COLLOC_USAGE_THRESH),
            weight_cpu_use_burst=config_manager.get(WEIGHT_CPU_USE_BURST, DEFAULT_WEIGHT_CPU_USE_BURST))

        self.__solver_max_runtime_secs = solver_max_runtime_secs
        self.__cpu_usage_predictor_manager = cpu_usage_predictor_manager
        self.__config_manager = config_manager
        self.__workload_monitor_manager = workload_monitor_manager
        self.__cnt_rebalance_calls = 0

    def assign_threads(self, cpu: Cpu, workload_id: str, workloads: dict) -> Cpu:
        curr_ids_per_workload = cpu.get_workload_ids_to_thread_ids()

        return self.__place_threads(cpu, workload_id, workloads, curr_ids_per_workload, True)

    def free_threads(self, cpu: Cpu, workload_id: str, workloads: dict) -> Cpu:
        curr_ids_per_workload = cpu.get_workload_ids_to_thread_ids()

        if workload_id not in curr_ids_per_workload:
            raise Exception("workload_id=`%s` is not placed on the instance. Cannot free it." % (workload_id,))

        return self.__place_threads(cpu, workload_id, workloads, curr_ids_per_workload, False)

    def rebalance(self, cpu: Cpu, workloads: dict) -> Cpu:
        self.__cnt_rebalance_calls += 1

        if len(workloads) == 0:
            return cpu

        log.info("Rebalancing with predictions...")
        # slow path, predict and adjust
        curr_ids_per_workload = cpu.get_workload_ids_to_thread_ids()

        try:
            return self.__place_threads(cpu, None, workloads, curr_ids_per_workload, None)
        except:
            log.error("Failed to rebalance, doing nothing.")
            self.__rebalance_failure_count += 1
            return cpu

    def __get_cpu_usage_predictor(self):
        return self.__cpu_usage_predictor_manager.get_predictor()

    def __predict_usage(self, workloads) -> dict:
        res_static = {}
        res_burst = {}
        cpu_usage_predictor = self.__get_cpu_usage_predictor()

        wmm = self.__workload_monitor_manager
        cm = self.__config_manager
        pred_env = PredEnvironment(cm.get_region(), cm.get_environment(), dt.utcnow().hour)

        cpu_usages = wmm.get_cpu_usage(seconds=3600, agg_granularity_secs=60)

        for w in workloads.values():  # TODO: batch the call
            pred = cpu_usage_predictor.predict(w, cpu_usages.get(w.get_id(), None), pred_env)
            if w.get_type() == STATIC:
                res_static[w.get_id()] = pred
            elif w.get_type() == BURST:
                res_burst[w.get_id()] = pred
        log.info("Usage prediction per static workload: " + str(res_static))
        log.info("Usage prediction per burst workload: " + str(res_burst))
        return res_static, res_burst

    def __place_threads(self, cpu, workload_id, workloads, curr_ids_per_workload, is_add):
        # this will predict against the new or deleted workload too if it's static
        predicted_usage_static, predicted_usage_burst = self.__predict_usage(workloads)
        cu_vector = self.__get_requested_cu_vector(cpu, workload_id, workloads, curr_ids_per_workload, is_add)

        # TODO: (maybe?) add burst pool current placement to curr_placement_vectors_static

        log.debug("workloads: {}".format(list(workloads.keys())))
        log.debug("cu_vector: {}".format(cu_vector))
        log.debug("predicted_usage_static: {}".format(predicted_usage_static))
        log.debug("predicted_usage_burst: {}".format(predicted_usage_burst))

        cpu = self.__compute_apply_placement(
            cpu,
            cu_vector.requested_cus,
            cu_vector.burst_pool_size_req,
            cu_vector.curr_placement_vectors_static,
            predicted_usage_static,
            predicted_usage_burst,
            workloads,
            cu_vector.ordered_workload_ids_static,
            cu_vector.ordered_workload_ids_burst)

        return cpu

    @staticmethod
    def __get_requested_cu_vector(cpu, workload_id, workloads, curr_ids_per_workload, is_add) -> CUVector:
        n_compute_units = len(cpu.get_threads())
        tid_2order = cpu.get_natural_indexing_2_original_indexing()

        ordered_workload_ids_static = []
        ordered_workload_ids_burst = []
        for w in get_sorted_workloads(workloads.values()):
            if w.get_type() == STATIC:
                ordered_workload_ids_static.append(w.get_id())
            else:
                ordered_workload_ids_burst.append(w.get_id())

        changed_workload = workloads.get(workload_id, None)

        curr_placement_vectors_static = []
        for wid in ordered_workload_ids_static:
            if (changed_workload is not None) and (wid == changed_workload.get_id()) and is_add:
                continue
            cids = curr_ids_per_workload[wid]
            v = [1 if tid_2order[i] in cids else 0 for i in range(n_compute_units)]
            curr_placement_vectors_static.append(v)
        if len(curr_placement_vectors_static) == 0:
            curr_placement_vectors_static = None
            requested_cus = []
        else:
            if (changed_workload is not None) and (changed_workload.get_type() == STATIC) and (not is_add):
                requested_cus = [len(curr_ids_per_workload[wid]) if wid != changed_workload.get_id() else 0 for wid in
                                 ordered_workload_ids_static]
            elif changed_workload is not None:
                requested_cus = [len(curr_ids_per_workload[wid]) for wid in ordered_workload_ids_static if
                                 wid != changed_workload.get_id()]
            else:
                requested_cus = [len(curr_ids_per_workload[wid]) for wid in ordered_workload_ids_static]

        if (changed_workload is not None) and (changed_workload.get_type() == STATIC) and is_add:
            requested_cus += [changed_workload.get_thread_count()]

        burst_workloads = get_burst_workloads(workloads.values())
        burst_pool_size_req = sum([w.get_thread_count() for w in burst_workloads]) if len(burst_workloads) > 0 else 0

        return CUVector(
            requested_cus,
            curr_placement_vectors_static,
            ordered_workload_ids_static,
            ordered_workload_ids_burst,
            burst_pool_size_req)

    @staticmethod
    def __assign_new_mapping(cpu, thread_id2workload_ids):
        cpu.clear()
        for t in cpu.get_threads():
            wids = thread_id2workload_ids.get(t.get_id(), None)
            if wids is not None:
                for wi in wids:
                    t.claim(wi)

    def __compute_apply_placement(
            self,
            cpu,
            requested_cus,
            burst_pool_size_req,
            curr_placement_vectors_static,
            predicted_usage_static,
            predicted_usage_burst,
            workloads,
            ordered_workload_ids_static,
            ordered_workload_ids_burst):

        predicted_usage_static_vector = None
        if len(predicted_usage_static) > 0:
            predicted_usage_static_vector = [predicted_usage_static[w_id] for w_id in ordered_workload_ids_static]

        sum_burst_pred = 0.0
        if len(predicted_usage_burst) > 0:
            sum_burst_pred = sum(v for v in predicted_usage_burst.values())

        new_placement_vectors = self.__compute_new_placement(
            cpu,
            requested_cus,
            burst_pool_size_req,
            sum_burst_pred,
            curr_placement_vectors_static,
            predicted_usage_static_vector)

        new_placement_vectors_static = new_placement_vectors[:-1] if burst_pool_size_req != 0 else new_placement_vectors
        new_placement_vector_burst = new_placement_vectors[-1] if burst_pool_size_req != 0 else None

        tid_2order = cpu.get_natural_indexing_2_original_indexing()
        thread_id2workload_ids = defaultdict(list)

        for w_ind, v in enumerate(new_placement_vectors_static):
            for i, e in enumerate(v):
                if e == 1:
                    thread_id2workload_ids[tid_2order[i]].append(ordered_workload_ids_static[w_ind])

        if new_placement_vector_burst is not None:
            for i, e in enumerate(new_placement_vector_burst):
                if e == 1:
                    for w_ind in ordered_workload_ids_burst:
                        thread_id2workload_ids[tid_2order[i]].append(w_ind)

        release_all_threads(cpu, workloads.values())
        self.__assign_new_mapping(cpu, thread_id2workload_ids)

        # TODO: log what's in print_statistics of compute_v2
        return cpu

    def __compute_new_placement(
            self,
            cpu,
            requested_units,
            burst_pool_size_req,
            sum_burst_pred,
            current_placement,
            predicted_usage_static):

        num_threads = len(cpu.get_threads())

        ip_params = self.__ip_solver_params
        if burst_pool_size_req > 0 and sum_burst_pred > 0:
            ip_params.weight_cpu_use_burst = sum_burst_pred / burst_pool_size_req

        # if the instance is mostly empty, loosen constraint
        # on burst pool growth size
        num_req_static = sum(requested_units)
        if num_req_static < num_threads / 2 and burst_pool_size_req > 0:
            ip_params = copy.deepcopy(ip_params)
            th_num_cores_static = ceil(num_req_static)
            ip_params.max_burst_pool_increase_ratio = floor((num_threads - th_num_cores_static) / burst_pool_size_req)

        placement, status = optimize_ip(
            requested_units,
            burst_pool_size_req,
            num_threads,
            len(cpu.get_packages()),
            previous_allocation=current_placement,
            use_per_workload=predicted_usage_static,
            solver_params=ip_params,
            verbose=False,
            max_runtime_secs=self.__solver_max_runtime_secs)

        if status == IP_SOLUTION_TIME_BOUND:
            self.__time_bound_call_count += 1

        return placement

    def set_solver_max_runtime_secs(self, val):
        self.__solver_max_runtime_secs = val

    def set_registry(self, registry):
        self.__reg = registry

    def report_metrics(self, tags):
        self.__reg.gauge(IP_ALLOCATOR_TIMEBOUND_COUNT, tags).set(self.__time_bound_call_count)
        self.__reg.gauge(FORECAST_REBALANCE_FAILURE_COUNT, tags).set(self.__rebalance_failure_count)



import datetime

from titus_isolate import log
from titus_isolate.allocate.fall_back_cpu_allocator import FallbackCpuAllocator
from titus_isolate.allocate.greedy_cpu_allocator import GreedyCpuAllocator
from titus_isolate.allocate.integer_program_cpu_allocator import IntegerProgramCpuAllocator
from titus_isolate.allocate.forecast_ip_cpu_allocator import ForecastIPCpuAllocator
from titus_isolate.allocate.noop_allocator import NoopCpuAllocator
from titus_isolate.allocate.noop_reset_allocator import NoopResetCpuAllocator
from titus_isolate.allocate.remote_cpu_allocator import RemoteCpuAllocator
from titus_isolate.config.config_manager import ConfigManager
from titus_isolate.config.constants import CPU_ALLOCATOR, CPU_ALLOCATORS, DEFAULT_ALLOCATOR, \
    CPU_ALLOCATOR_A, CPU_ALLOCATOR_B, AB_TEST, EC2_INSTANCE_ID, IP, GREEDY, NOOP, FORECAST_CPU_IP, \
    NOOP_RESET, FREE_THREAD_PROVIDER, DEFAULT_FREE_THREAD_PROVIDER, EMPTY, THRESHOLD, DEFAULT_TOTAL_THRESHOLD, \
    TOTAL_THRESHOLD, DEFAULT_THRESHOLD_TOTAL_DURATION_SEC, THRESHOLD_TOTAL_DURATION_SEC, DEFAULT_PER_WORKLOAD_THRESHOLD, \
    PER_WORKLOAD_THRESHOLD, DEFAULT_PER_WORKLOAD_DURATION_SEC, PER_WORKLOAD_DURATION_SEC, FALLBACK_ALLOCATOR, \
    DEFAULT_FALLBACK_ALLOCATOR, REMOTE
from titus_isolate.monitor.empty_free_thread_provider import EmptyFreeThreadProvider
from titus_isolate.monitor.free_thread_provider import FreeThreadProvider
from titus_isolate.monitor.threshold_free_thread_provider import ThresholdFreeThreadProvider
from titus_isolate.utils import get_config_manager, get_cpu_usage_predictor_manager, get_workload_monitor_manager

BUCKETS = ["A", "B"]

CPU_ALLOCATOR_NAME_TO_CLASS_MAP = {
    IP: IntegerProgramCpuAllocator,
    GREEDY: GreedyCpuAllocator,
    NOOP: NoopCpuAllocator,
    NOOP_RESET: NoopResetCpuAllocator,
    REMOTE: RemoteCpuAllocator
}


def get_free_thread_provider(config_manager: ConfigManager) -> FreeThreadProvider:
    free_thread_provider_str = config_manager.get(FREE_THREAD_PROVIDER, DEFAULT_FREE_THREAD_PROVIDER)
    free_thread_provider = None

    if free_thread_provider_str == EMPTY:
        free_thread_provider = EmptyFreeThreadProvider()
    elif free_thread_provider_str == THRESHOLD:
        total_threshold = config_manager.get(TOTAL_THRESHOLD, DEFAULT_TOTAL_THRESHOLD)
        total_duration_sec = config_manager.get(THRESHOLD_TOTAL_DURATION_SEC, DEFAULT_THRESHOLD_TOTAL_DURATION_SEC)
        per_workload_threshold = config_manager.get(PER_WORKLOAD_THRESHOLD, DEFAULT_PER_WORKLOAD_THRESHOLD)
        per_workload_duration_sec = config_manager.get(PER_WORKLOAD_DURATION_SEC, DEFAULT_PER_WORKLOAD_DURATION_SEC)

        free_thread_provider = ThresholdFreeThreadProvider(
            total_threshold=total_threshold,
            total_duration_sec=total_duration_sec,
            per_workload_threshold=per_workload_threshold,
            per_workload_duration_sec=per_workload_duration_sec)

    log.debug("Free thread provider: '{}'".format(free_thread_provider.__class__.__name__))
    return free_thread_provider


def get_allocator(config_manager, hour=None) -> FallbackCpuAllocator:
    if hour is None:
        hour = datetime.datetime.utcnow().hour

    primary_alloc_str = config_manager.get(CPU_ALLOCATOR)
    secondary_alloc_str = config_manager.get(FALLBACK_ALLOCATOR, DEFAULT_FALLBACK_ALLOCATOR)

    if primary_alloc_str == AB_TEST:
        primary_allocator = __get_ab_allocator(config_manager, hour)
    else:
        primary_allocator = __get_allocator(primary_alloc_str, config_manager)

    secondary_allocator = __get_allocator(secondary_alloc_str, config_manager)

    return FallbackCpuAllocator(primary_allocator, secondary_allocator)


def __get_allocator(allocator_str, config_manager):
    if allocator_str not in CPU_ALLOCATORS:
        log.error("Unexpected CPU allocator specified: '{}', falling back to default: '{}'".format(allocator_str, DEFAULT_ALLOCATOR))
        allocator_str = DEFAULT_ALLOCATOR

    if allocator_str != FORECAST_CPU_IP:
        free_thread_provider = get_free_thread_provider(config_manager)
        return CPU_ALLOCATOR_NAME_TO_CLASS_MAP[allocator_str](free_thread_provider)

    return ForecastIPCpuAllocator(
        cpu_usage_predictor_manager=get_cpu_usage_predictor_manager(),
        config_manager=get_config_manager(),
        workload_monitor_manager=get_workload_monitor_manager())


def __get_ab_allocator(config_manager, hour):
    a_allocator_str = config_manager.get(CPU_ALLOCATOR_A)
    b_allocator_str = config_manager.get(CPU_ALLOCATOR_B)

    a_allocator = __get_allocator(a_allocator_str, config_manager)
    b_allocator = __get_allocator(b_allocator_str, config_manager)

    bucket = get_ab_bucket(config_manager, hour)

    if bucket not in BUCKETS:
        log.error(
            "Unexpected A/B bucket specified: '{}', falling back to default: '{}'".format(bucket, DEFAULT_ALLOCATOR))
        return __get_allocator("UNDEFINED_AB_BUCKET", config_manager)

    return {
        "A": a_allocator,
        "B": b_allocator,
    }[bucket]


def get_ab_bucket(config_manager, hour):
    instance_id = config_manager.get(EC2_INSTANCE_ID)
    if instance_id is None:
        log.error("Failed to find: '{}' in config manager, is the environment variable set?".format(EC2_INSTANCE_ID))
        return "UNDEFINED"

    # Take the last character of an instance id turn it into a number and determine whether it's odd or even.
    # An instance id looks like this: i-0cfefd19c9a8db976
    char = instance_id[-1]
    bucket = _get_ab_bucket_int(char, hour)
    if bucket == 0:
        return "A"
    else:
        return "B"


def _get_ab_bucket_int(char, hour):
    instance_int = ord(char)
    hour = int(hour / 6)
    int_bucket = (instance_int + (hour % 2)) % 2
    return int_bucket



import datetime
import os
import socket
import uuid

import requests

from titus_isolate import log
from titus_isolate.config.constants import EVENT_LOG_FORMAT_STR
from titus_isolate.model.processor.cpu import Cpu
from titus_isolate.utils import get_config_manager, get_workload_monitor_manager


def get_address(region, env, stream):
    format_str = get_config_manager().get(EVENT_LOG_FORMAT_STR, None)
    if format_str is None:
        return None

    return format_str.format(region, env, stream)


def send_event_msg(msg, address):
    log.debug("Sending to keystone address: '{}' msg: '{}'".format(address, msg))
    r = requests.post(address, json=msg)
    log.debug("Received response: '{}'".format(r))


def get_event_msg(event):
    app_name = "titus-isolate"
    host_name = socket.gethostname()
    ack = True

    return {
        "appName": app_name,
        "hostname": host_name,
        "ack": ack,
        "event": [event]
    }


def get_cpu_event(cpu: Cpu, usage: dict, workloads: dict):
    return {
        "uuid": str(uuid.uuid4()),
        "payload": {
            "ts": str(datetime.datetime.utcnow()),
            "instance": os.environ['EC2_INSTANCE_ID'],
            "cpu": cpu.to_dict(),
            "cpu_usage": usage,
            "workloads:": workloads
        }
    }


def report_cpu(cpu: Cpu, workloads):
    address = __get_address()
    if address is None:
        return

    workload_monitor_manager = get_workload_monitor_manager()
    if workload_monitor_manager is None:
        log.debug("Failed to retrieve workload monitor manager to report cpu.")
        return

    usage = workload_monitor_manager.get_cpu_usage(600, 60)

    serializable_usage = {}
    for w_id, usage in usage.items():
        serializable_usage[w_id] = [str(u) for u in usage]

    serializable_workloads = {}
    for w in workloads:
        serializable_workloads[w.get_id()] = w.to_dict()

    msg = get_event_msg(get_cpu_event(cpu, serializable_usage, serializable_workloads))
    log.debug("reporting cpu: {}".format(msg))
    send_event_msg(msg, address)


def __get_address():
    config_manager = get_config_manager()
    region = config_manager.get_region()
    env = config_manager.get_environment()
    stream = 'titus_isolate'

    address = get_address(region, env, stream)
    if address is None:
        log.error("Failed to retrieve event log address for region: '{}', env: '{}', stream: '{}'".format(
            region, env, stream))
        return None

    return address



from functools import reduce

from titus_isolate.model.processor.utils import get_packages_with_workload, get_workload_ids


def get_cross_package_violations(cpu):
    """
    Returns a dictionary mapping workload ids to lists of package ids.  Only workloads on more than one package are
    included.

    {
       fa873f01-da52-45b4-b37b-edad0dfab519: []
    }

    :param cpu: CPU to scan for cross package violations
    :return: dictionary mapping workload ids to lists of packages
    """
    violations = {}
    workload_ids = get_workload_ids(cpu)

    for workload_id in workload_ids:
        packages = get_packages_with_workload(cpu, workload_id)
        if len(packages) > 1:
            violations[workload_id] = [p.get_id() for p in packages]

    return violations


def get_shared_core_violations(cpu):
    """
    Returns a dictionary mapping core ids to lists of workload ids.  Only workloads on more than one package are
    included.

    {
       12: [c520b163-daec-4c54-a835-585ba05915ca , c8c23650-d37f-4270-b36e-b74d27dfa709],
       <core_id>: [<workload_id>, <workload_id>]
       ...
    }

    :param cpu: CPU to scan for cross package violations
    :return: dictionary mapping core ids to lists of workload ids
    """
    violations = {}

    for package in cpu.get_packages():
        for core in package.get_cores():
            unique_workload_ids = set(__get_workload_ids(core))
            if len(unique_workload_ids) > 1:
                violation_key = ':'.join([str(package.get_id()), str(core.get_id())])
                violations[violation_key] = list(unique_workload_ids)

    return violations


def __get_workload_ids(core):
    workload_ids = []
    for t in core.get_threads():
        workload_ids += t.get_workload_ids()

    return workload_ids



from functools import reduce

from titus_isolate.model.processor.thread import Thread

DEFAULT_PACKAGE_COUNT = 2
DEFAULT_CORE_COUNT = 4
DEFAULT_THREAD_COUNT = 2
DEFAULT_TOTAL_THREAD_COUNT = DEFAULT_PACKAGE_COUNT * DEFAULT_CORE_COUNT * DEFAULT_THREAD_COUNT


def get_empty_threads(threads):
    return [t for t in threads if not t.is_claimed()]


def get_claimed_threads(threads):
    return [t for t in threads if t.is_claimed()]


def get_emptiest_core(package):
    emptiest_core = package.get_cores()[0]
    curr_empty_thread_count = len(emptiest_core.get_empty_threads())

    for core in package.get_cores()[1:]:
        new_empty_thread_count = len(core.get_empty_threads())
        if new_empty_thread_count > curr_empty_thread_count:
            emptiest_core = core
            curr_empty_thread_count = new_empty_thread_count

    return emptiest_core


def is_cpu_full(cpu):
    empty_threads = reduce(list.__add__, [p.get_empty_threads() for p in cpu.get_packages()])
    return len(empty_threads) == 0


# Workloads
def get_workload_ids(cpu):
    workload_ids = []
    for t in cpu.get_threads():
        workload_ids += t.get_workload_ids()

    return set(workload_ids)


def get_packages_with_workload(cpu, workload_id):
    return [package for package in cpu.get_packages() if is_on_package(package, workload_id)]


def is_on_package(package, workload_id):
    return len(get_threads_with_workload(package, workload_id)) > 0


def get_threads_with_workload(core, workload_id):
    return [thread for thread in core.get_threads() if workload_id in thread.get_workload_ids()]


def __get_str_repr(ind):
    res = ''
    while ind >= 0:
        res += chr(ord('a') + (ind % 26))
        ind -= 26
    return res


def __get_workloads_str(t: Thread) -> str:
    return str(sorted([str(w_id) for w_id in t.get_workload_ids()]))


def __get_footprints(cpus) -> dict:
    footprints = {}
    keys = []
    for cpu in cpus:
        for t in cpu.get_threads():
            if not t.is_claimed():
                continue

            workloads_str = __get_workloads_str(t)
            if workloads_str not in keys:
                keys.append(workloads_str)

    for i, k in enumerate(keys):
        footprints[k] = __get_str_repr(i)
        i += 1

    return footprints


def visualize_cpu_comparison(old_cpu, new_cpu):
    footprints = __get_footprints([old_cpu, new_cpu])
    return "\n{}\n{}".format(
        __visualize(old_cpu, footprints),
        __visualize(new_cpu, footprints))


def visualize(cpu):
    footprints = __get_footprints([cpu])
    return __visualize(cpu, footprints)


def __visualize(cpu, footprints):
    """
    This function will return a string representing the current layout of
    workloads on the given cpu. Example:

    | a | c | b | b |   a: job1
    | b | c | d | d |   b: job2
    | ------------- |   c: job3
    | e | e | e | e |   d: job4
    | e | e | e | e |   e: job5

    In this example, there are 5 workloads, layed out on 2 sockets, each with 4 physical cores.
    Workload with id=`job3` is using the 2 hyperthreads of the 2nd core of the first socket.
    """
    buffer = []

    n_packages = len(cpu.get_packages())
    n_compute_units = len(cpu.get_threads())
    n_cus_per_package = n_compute_units // n_packages

    for ind_p, package in enumerate(cpu.get_packages()):
        S1 = [' '] * (n_cus_per_package // 2)
        S2 = [' '] * (n_cus_per_package // 2)

        j = 0
        for core in package.get_cores():
            threads = core.get_threads()
            for t in threads:
                if not t.is_claimed():
                    j += 1
                    continue

                simple_id = footprints[__get_workloads_str(t)]
                if j % 2 == 0:
                    S1[j // 2] = simple_id
                else:
                    S2[j // 2] = simple_id
                j +=1
        buffer.append('| ' + ' | '.join(S1) + ' |\n')
        buffer.append('| ' + ' | '.join(S2) + ' |\n')
        if ind_p < n_packages - 1:
            buffer.append('| ' + '-' * len(' | '.join(S1)) + ' |\n')
    lgrid = len(buffer[0])

    workloads_on_cpu = [__get_workloads_str(t) for t in cpu.get_threads()]
    i = 0
    for wid, simple_id in footprints.items():
        if wid not in workloads_on_cpu:
            continue
        if i < len(buffer):
            buffer[i] = buffer[i][:-1] + '   %s: %s\n' % (simple_id, wid)
        else:
            buffer.append(' ' * lgrid + '  %s: %s\n' % (simple_id, wid))
        i += 1
    return ''.join(buffer)



from titus_isolate.model.processor import utils


class Core:
    def __init__(self, identifier, threads):
        if len(threads) < 1:
            raise ValueError("A CPU core must have at least 1 thread")

        self.__identifier = identifier
        self.__threads = threads

    def get_id(self):
        return self.__identifier

    def get_threads(self):
        return self.__threads

    def get_empty_threads(self):
        return utils.get_empty_threads(self.get_threads())



import copy

from titus_isolate.event.constants import ACTION, RECONCILE
from titus_isolate.event.event_handler import EventHandler
from titus_isolate.isolate.reconciler import Reconciler
from titus_isolate.utils import get_workload_manager


class ReconcileEventHandler(EventHandler):
    def __init__(self, reconciler: Reconciler):
        super().__init__(None)
        self.__reconciler = reconciler

    def handle(self, event):
        if not self.__relevant(event):
            return

        cpu = copy.deepcopy(get_workload_manager().get_cpu())
        self.handling_event(event, "reconciling titus-isolate and cgroup state")
        self.__reconciler.reconcile(cpu)
        self.handled_event(event, "reconciled titus-isolate and cgroup state")

    def __relevant(self, event):
        if not event[ACTION] == RECONCILE:
            self.ignored_event(event, "not a {} event".format(RECONCILE))
            return False

        return True



from titus_isolate import log
from titus_isolate.utils import get_config_manager


def get_required_property(key):
    value = get_config_manager().get(key)
    if value is None:
        log.error("Failed to retrieve property: '{}'".format(key))
        return None

    return value



from titus_isolate.isolate.detect import get_cross_package_violations, get_shared_core_violations


def has_better_isolation(cur_cpu, new_cpu):
    """
    Here we determine whether a proposed placement of workloads improves upon the current workload placement.

    There are two possible violations:
       1. shared_core: a core is shared by multiple workloads
       2. cross_package: a workload is placed on multiple packages

    Below we describe a matrix answering the follow question: The new workload placement is better?

    There are 9 possible cases to consider.
        '-' indicates a decrease
        '0' indicates no change
        '+' indicates an increase
        'count' refers to how many workloads are affected by the violation



                                    shared_core_count
                                    -      0      +

                                -   T      T      T


            cross_package_count 0   T      F      F


                                +   F      F      F


        Only two controversial answers exist.  They are in the lower-left and upper-right hand corners.

            lower-left: shared_core_count has decreased, but cross_package_count has increased
            upper-right: cross_package_count has decreased, but shared_core_count has increased

        In both cases we make the assumption that avoiding cross package workloads is to be preferred.

    NOTE: We do not consider the placement of burst workloads when comparing the two placement options.

    :return: True if the new_cpu has better placement, False otherwise
    """
    cur_cross_package_violation_count = len(get_cross_package_violations(cur_cpu))
    new_cross_package_violation_count = len(get_cross_package_violations(new_cpu))

    cur_shared_core_violation_count = len(get_shared_core_violations(cur_cpu))
    new_shared_core_violation_count = len(get_shared_core_violations(new_cpu))

    # More violations is bad, so a positive change is bad
    cross_package_violation_change = new_cross_package_violation_count - cur_cross_package_violation_count
    shared_core_violation_change = new_shared_core_violation_count - cur_shared_core_violation_count

    # Top row of matrix
    if cross_package_violation_change < 0:
        return True

    # Bottom row of matrix
    if cross_package_violation_change > 0:
        return False

    # Middle row of matrix, can assume cross_package_violation_change == 0
    return shared_core_violation_change < 0




import datetime

from titus_isolate import log
from titus_isolate.cgroup.utils import parse_cpuacct_usage_all, get_usage_all_path
from titus_isolate.monitor.cpu_usage import CpuUsageSnapshot


class CgroupMetricsProvider:

    def __init__(self, workload):
        self.__workload = workload
        self.__usage_path = None

    def get_workload(self):
        return self.__workload

    def get_cpu_usage(self):
        usage_path = self.__get_usage_path()
        if usage_path is None:
            return None

        with open(usage_path, 'r') as f:
            timestamp = datetime.datetime.utcnow()
            content = f.read()

        cpu_usage_rows = parse_cpuacct_usage_all(content)
        return CpuUsageSnapshot(timestamp, cpu_usage_rows)

    def __get_usage_path(self):
        if self.__usage_path is not None:
            return self.__usage_path

        try:
            self.__usage_path = get_usage_all_path(self.__workload.get_id())
        except FileNotFoundError:
            log.warning("No cpu usage path for workload: '{}'".format(self.__workload.get_id()))

        return self.__usage_path



import logging
from threading import Lock

from flask import Flask, request, jsonify

from titus_isolate import log
from titus_isolate.allocate.utils import parse_workload, parse_cpu
from titus_isolate.config.env_property_provider import EnvPropertyProvider
from titus_isolate.isolate.utils import get_allocator
from titus_isolate.utils import get_config_manager

lock = Lock()
cpu_allocator = None

app = Flask(__name__)

if __name__ != '__main__':
    gunicorn_logger = logging.getLogger('gunicorn.error')
    app.logger.handlers = gunicorn_logger.handlers
    app.logger.setLevel(gunicorn_logger.level)


def get_cpu_allocator():
    with lock:
        global cpu_allocator
        if cpu_allocator is None:
            config_manager = get_config_manager(EnvPropertyProvider())
            cpu_allocator = get_allocator(config_manager)

        return cpu_allocator


def set_cpu_allocator(allocator):
    with lock:
        global cpu_allocator
        cpu_allocator = allocator


def __get_cpu(body):
    return parse_cpu(body["cpu"])


def __get_workload_id(body):
    return body["workload_id"]


def __get_workloads(body):
    workloads = {}
    for w in [parse_workload(w_str) for w_str in body["workloads"]]:
        workloads[w.get_id()] = w

    return workloads


def get_threads_arguments(body):
    workload_id = __get_workload_id(body)
    cpu = __get_cpu(body)
    workloads = __get_workloads(body)
    return cpu, workload_id, workloads


def get_rebalance_arguments(body):
    cpu = __get_cpu(body)
    workloads = __get_workloads(body)
    return cpu, workloads


@app.route('/cpu_allocator', methods=['GET'])
def get_allocator():
    allocator = get_cpu_allocator()
    if cpu_allocator is None:
        return "CPU allocator not set", 404

    return allocator.__class__.__name__


@app.route('/assign_threads', methods=['PUT'])
def assign_threads():
    allocator = get_cpu_allocator()

    if allocator is None:
        log.error("Cannot assign threads, CPU allocator is not set.")
        return "CPU allocator not set", 500

    try:
        body = request.get_json()
        cpu_in, workload_id, workloads = get_threads_arguments(body)
        cpu_out = allocator.assign_threads(cpu_in, workload_id, workloads)
        return jsonify(cpu_out.to_dict())
    except:
        log.exception("Failed to assign threads")
        return "Failed to assign threads", 500


@app.route('/free_threads', methods=['PUT'])
def free_threads():
    allocator = get_cpu_allocator()

    if allocator is None:
        log.error("Cannot free threads, CPU allocator is not set.")
        return "CPU allocator not set", 500

    try:
        body = request.get_json()
        cpu_in, workload_id, workloads = get_threads_arguments(body)
        cpu_out = allocator.free_threads(cpu_in, workload_id, workloads)
        return jsonify(cpu_out.to_dict())
    except:
        log.exception("Failed to free threads")
        return "Failed to free threads", 500


@app.route('/rebalance', methods=['PUT'])
def rebalance():
    allocator = get_cpu_allocator()

    if allocator is None:
        log.error("Cannot rebalance threads, CPU allocator is not set.")
        return "CPU allocator not set", 500

    try:
        body = request.get_json()
        cpu_in, workloads = get_rebalance_arguments(body)
        cpu_out = allocator.rebalance(cpu_in, workloads)
        return jsonify(cpu_out.to_dict())
    except:
        log.exception("Failed to rebalance")
        return "Failed to rebalance", 500





import logging
from abc import abstractmethod

from titus_isolate import log


class EventHandler:
    def __init__(self, workload_manager, log_level=logging.INFO):
        self.workload_manager = workload_manager
        self.__log_level = log_level
        self.__ignored_event_count = 0
        self.__handled_event_count = 0

    def ignored_event(self, event, msg):
        log.debug("'{}' ignored event. msg: '{}', event: '{}', ".format(self.__class__.__name__, msg, event))
        self.__ignored_event_count += 1

    def handling_event(self, event, msg):
        log.log(self.__log_level, "'{}' handling event.  msg: '{}', event: '{}'".format(self.__class__.__name__, msg, event))

    def handled_event(self, event, msg):
        log.log(self.__log_level, "'{}' handled event.  msg: '{}', event: '{}'".format(self.__class__.__name__, msg, event))
        self.__handled_event_count += 1

    def get_ignored_event_count(self):
        return self.__ignored_event_count

    def get_handled_event_count(self):
        return self.__handled_event_count

    @abstractmethod
    def handle(self, event):
        pass




from functools import reduce

from titus_isolate.model.processor import utils


class Package:
    def __init__(self, identifier, cores):
        if len(cores) < 1:
            raise ValueError("A CPU package must have at least 1 core.")

        self.__identifier = identifier
        self.__cores = cores

    def get_id(self):
        return self.__identifier

    def get_cores(self):
        return self.__cores

    def get_threads(self):
        return reduce(list.__add__, [core.get_threads() for core in self.get_cores()])

    def get_empty_threads(self):
        return utils.get_empty_threads(self.get_threads())
